/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * include/SIL/random.h: Header for random number generation routines.
 */

/*
 * This header declares an interface to a linear congruential pseudo-random
 * number generator whose behavior is consistent for a given seed value
 * regardless of the runtime environment.  The exact generator formula is:
 *
 *     x[n+1] = ((x[n] * 6364136223846793005) + 1) mod 2^64
 *
 * where x[n] is the internal generator state at iteration n.  The actual
 * value returned to the caller is the upper B bits of the new state, where
 * B is the bit width of the value to be returned (for floating point
 * functions, B can vary depending on the floating point data format and
 * the value itself).  This generator is believed to be of reasonably good,
 * though not cryptographic, quality; in particular, urandom() passes the
 * dieharder (http://www.phy.duke.edu/~rgb/General/dieharder.php) tests as
 * of dieharder version 3.31.1.
 *
 * Note that the low-end bits of random64() and urandom64() are less random
 * than the high-end bits.  If you need a 64-bit value with high randomness
 * in all bits, concatenate two 32-bit random numbers instead.
 *
 * Entropy can be injected into the random number stream by, for example:
 *    - Extracting a new random number from the stream at intervals defined
 *      by external events, such as when an input event is received.
 *    - Modifying the random number seed based on external data, such as
 *      the current state of input devices.  (For the non-reentrant
 *      interfaces, this entails retrieving the seed with urandom64(),
 *      modifying the seed, then applying it with srandom64().)
 *
 * The random() and srandom() interfaces comply with the specifications of
 * rand() and srand(), respectively, in the C99 standard.  Thus, one can
 * redefine the standard functions as follows:
 *     #define rand      random
 *     #define srand     srandom
 *     #define RAND_MAX  RANDOM_MAX
 * and still get expected behavior.  The random_r() interface differs
 * slightly from the POSIX-specific rand_r() function in that random_r()
 * takes a 64-bit state pointer rather than an "unsigned int" state
 * pointer.
 *
 * The random() and srandom() functions are _not_ compatible with the
 * same-named functions in POSIX.1-2001, though code which does not call
 * initstate() or setstate() and which does not depend on a specific type
 * of generator should still work as designed.
 */

#ifndef SIL_RANDOM_H
#define SIL_RANDOM_H

EXTERN_C_BEGIN

#if defined(__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
# include <float.h>  // For FLT_RADIX, etc.
#endif

/*************************************************************************/
/*************************************************************************/

/**
 * RANDOM_MAX:  The maximum value returned by random().  Equal to INT_MAX
 * for a 32-bit int type.
 */
#define RANDOM_MAX  0x7FFFFFFF

/*-----------------------------------------------------------------------*/

/**
 * random():  Return a pseudo-random integer from 0 to 0x7FFF'FFFF, inclusive.
 *
 * This function is not thread-safe.  For thread-safe behavior, call
 * random_r() instead.
 *
 * [Return value]
 *     Pseudo-random number.
 */
extern int random(void);

/**
 * urandom():  Return a pseudo-random integer from 0 to 0xFFFF'FFFF, inclusive.
 *
 * This function is not thread-safe.  For thread-safe behavior, call
 * urandom_r() instead.
 *
 * [Return value]
 *     Pseudo-random number.
 */
extern unsigned int urandom(void);

/**
 * random64():  Return a pseudo-random integer from 0 to
 * 0x7FFF'FFFF'FFFF'FFFF, inclusive.
 *
 * This function is not thread-safe.  For thread-safe behavior, call
 * random64_r() instead.
 *
 * [Return value]
 *     Pseudo-random number.
 */
extern int64_t random64(void);

/**
 * urandom64():  Return a pseudo-random integer from 0 to
 * 0xFFFF'FFFF'FFFF'FFFF, inclusive.  See urandom64_r() for notes on the
 * relationship between the value returned and the current random seed.
 *
 * This function is not thread-safe.  For thread-safe behavior, call
 * urandom64_r() instead.
 *
 * [Return value]
 *     Pseudo-random number.
 */
extern uint64_t urandom64(void);

/**
 * frandom():  Return a pseudo-random double-precision floating point
 * number in the range [0.0,1.0).
 *
 * This function is not thread-safe.  For thread-safe behavior, call
 * frandom_r() instead.
 *
 * [Return value]
 *     Pseudo-random number.
 */
extern double frandom(void);

/**
 * frandomf():  Return a pseudo-random single-precision floating point
 * number in the range [0.0,1.0).
 *
 * This function is not thread-safe.  For thread-safe behavior, call
 * frandomf_r() instead.
 *
 * [Return value]
 *     Pseudo-random number.
 */
extern float frandomf(void);

/*-----------------------------------------------------------------------*/

/* To avoid code duplication: */
#define RNG_FUNC(x)  (((x) * 6364136223846793005ULL) + 1)

/**
 * random_r():  Return a pseudo-random integer from 0 to 0x7FFF'FFFF,
 * inclusive.
 *
 * [Parameters]
 *     state: Pointer to random number generator state.
 * [Return value]
 *     Pseudo-random number.
 */
#if defined(__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
inline int random_r(uint64_t *state);  // Silence -Wmissing-declarations.
inline int random_r(uint64_t *state) {
    *state = RNG_FUNC(*state);
    return *state >> 33;
}
#else
extern int random_r(uint64_t *state);
#endif

/**
 * urandom_r():  Return a pseudo-random integer from 0 to 0xFFFF'FFFF,
 * inclusive.
 *
 * [Parameters]
 *     state: Pointer to random number generator state.
 * [Return value]
 *     Pseudo-random number.
 */
#if defined(__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
inline unsigned int urandom_r(uint64_t *state);
inline unsigned int urandom_r(uint64_t *state) {
    *state = RNG_FUNC(*state);
    return *state >> 32;
}
#else
extern unsigned int urandom_r(uint64_t *state);
#endif

/**
 * random64_r():  Return a pseudo-random integer from 0 to
 * 0x7FFF'FFFF'FFFF'FFFF, inclusive.
 *
 * [Parameters]
 *     state: Pointer to random number generator state.
 * [Return value]
 *     Pseudo-random number.
 */
#if defined(__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
inline int64_t random64_r(uint64_t *state);
inline int64_t random64_r(uint64_t *state) {
    *state = RNG_FUNC(*state);
    return *state >> 1;
}
#else
extern int64_t random64_r(uint64_t *state);
#endif

/**
 * urandom64_r():  Return a pseudo-random integer from 0 to
 * 0xFFFF'FFFF'FFFF'FFFF, inclusive.
 *
 * The value returned by this function is always equal to the value of the
 * state buffer after the function returns.  In other words, the following
 * assertions always hold (for (2), assuming that no other thread
 * concurrently calls one of the non-reentrant functions):
 *
 * (1) uint64_t state = <any 64-bit value>;
 *     uint64_t value = urandom64_r(&state);
 *     ASSERT(value == state);
 *
 * (2) uint64_t seed = urandom64();
 *     uint64_t value1 = urandom64();
 *     srandom64(seed);
 *     uint64_t value2 = urandom64();
 *     ASSERT(value2 == value1);
 *
 * [Parameters]
 *     state: Pointer to random number generator state.
 * [Return value]
 *     Pseudo-random number.
 */
#if defined(__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
inline uint64_t urandom64_r(uint64_t *state);
inline uint64_t urandom64_r(uint64_t *state) {
    *state = RNG_FUNC(*state);
    return *state;
}
#else
extern uint64_t urandom64_r(uint64_t *state);
#endif

/**
 * frandom_r():  Return a pseudo-random double-precision floating point
 * number in the range [0.0,1.0).
 *
 * [Parameters]
 *     state: Pointer to random number generator state.
 * [Return value]
 *     Pseudo-random number.
 */
#if defined(__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
inline double frandom_r(uint64_t *state);
inline double frandom_r(uint64_t *state) {
    *state = RNG_FUNC(*state);
    /* Care is needed here because simply converting the state value to
     * floating point and dividing by 2^64 will result in an output of 1.0
     * if the state value is greater than 2^64 * (1.0 - 0.5 ulp). */
    #if FLT_RADIX == 2
        #if DBL_MANT_DIG >= 64
             return (double)(*state) / ((double)(1ULL<<63) * 2);
        #else
             return (double)(*state >> (64 - DBL_MANT_DIG))
                    / (double)(1ULL << DBL_MANT_DIG);
        #endif
    #else
        double result = (double)(*state) / ((double)(1ULL<<63) * 2);
        if (UNLIKELY(result == 1.0)) {
            result = 0.0;
        }
        return result;
    #endif
}
#else
extern double frandom_r(uint64_t *state);
#endif

/**
 * frandomf_r():  Return a pseudo-random single-precision floating point
 * number in the range [0.0,1.0).
 *
 * [Parameters]
 *     state: Pointer to random number generator state.
 * [Return value]
 *     Pseudo-random number.
 */
#if defined(__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
inline float frandomf_r(uint64_t *state);
inline float frandomf_r(uint64_t *state) {
    *state = RNG_FUNC(*state);
    #if FLT_RADIX == 2
        #if FLT_MANT_DIG >= 64
             return (float)(*state) / ((float)(1ULL<<63) * 2);
        #else
             return (float)(*state >> (64 - FLT_MANT_DIG))
                    / (double)(1ULL << FLT_MANT_DIG);
        #endif
    #else
        float result = (float)(*state) / ((float)(1ULL<<63) * 2);
        if (UNLIKELY(result == 1.0)) {
            result = 0.0;
        }
        return result;
    #endif
}
#else
extern float frandomf_r(uint64_t *state);
#endif

#undef RNG_FUNC

/*-----------------------------------------------------------------------*/

/**
 * srandom():  Set the seed value for the random number generator used by
 * the random() family of functions based on the given int-sized value.
 *
 * If neither this function nor one of the other srandom*() functions is
 * called before obtaining random numbers, the program acts as though
 * srandom(1) had been called before the first random number was obtained.
 *
 * [Parameters]
 *     seed: Seed value.
 */
extern void srandom(unsigned int seed);

/**
 * srandom64():  Set the seed value for the random number generator used
 * the random() family of functions to the given 64-bit value.
 *
 * [Parameters]
 *     seed: Seed value.
 */
extern void srandom64(uint64_t seed);

/**
 * srandom_env():  Set the seed value for the random number generator used
 * the random() family of functions based on the current runtime environment.
 * The exact method used to set the seed is system-dependent, but typically
 * includes factors such as the current time of day.
 */
extern void srandom_env(void);

/*************************************************************************/
/*************************************************************************/

EXTERN_C_END

#endif  // SIL_RANDOM_H

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
