/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/macosx/view.h: Header for the SILView class.
 */

#ifndef SIL_SRC_SYSDEP_MACOSX_VIEW_H
#define SIL_SRC_SYSDEP_MACOSX_VIEW_H

#include "src/sysdep/macosx/osx-headers.h"
#include <AppKit/NSOpenGL.h>
#include <AppKit/NSView.h>
#include <OpenGL/OpenGL.h>

/*************************************************************************/
/*************************************************************************/

/**
 * SILView:  NSView subclass implementing OpenGL support.
 */

@interface SILView: NSView {
  @private
    NSOpenGLPixelFormat *pixel_format;
    NSOpenGLContext *context;
}

/**
 * initWithFrame:pixelFormat:  Initialize a SILView object with the given
 * size and OpenGL pixel format.
 *
 * [Parameters]
 *     frame: View size (frame.origin is ignored).
 *     format: OpenGL pixel format to use.
 * [Return value]
 *     The initialized SILView object, or nil on error.
 */
- (id)initWithFrame:(NSRect)frame pixelFormat:(NSOpenGLPixelFormat *)format;

/**
 * createGLContext:  Create (and set as current) an OpenGL context for the
 * current thread which is associated with this view.
 *
 * Currently, only one thread may have an OpenGL rendering context for a view.
 * Any number of non-rendering contexts may be created.
 *
 * [Parameters]
 *     forRendering: True to create a context suitable for rendering;
 *         false to create a context suitable only for shader compilation.
 * [Return value]
 *     True on success, false on error.
 */
- (BOOL)createGLContext:(BOOL)forRendering;

/**
 * destroyGLContext:  Destroy the OpenGL context (if any) for the current
 * thread.
 */
- (void)destroyGLContext;

@end

/*************************************************************************/
/*************************************************************************/

#endif  // SIL_SRC_SYSDEP_MACOSX_VIEW_H

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
